import sqlite3

from packaging.version import Version

from .constants import db_path
from app.__version__ import VERSION
from app.config import update_config
from app.constants import config_path


migration_list = [
    ("0.0.8", """ALTER TABLE tasks ADD COLUMN subtasks INTEGER DEFAULT 0;"""),
    ("0.0.12", """CREATE INDEX title on tasks(title);"""),
]


def run_migrations(previous_version):
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    previous_version = Version(previous_version)
    for migration in migration_list:
        if Version(migration[0]) > previous_version:
            cur.execute(migration[1])
            conn.commit()
    conn.close()


def update_version(config):
    if config["version"] != VERSION:
        run_migrations(previous_version=config["version"])
        config["version"] = VERSION
        update_config(config_path, config=config)



import datetime

from . import database
from app.utility import convert_to_console_date
from app.utility import convert_to_db_date
from app.utility import generate_migration_error
from app.utility import sanitize_text


def list_tasks(
    table="tasks",
    priority=None,
    today=None,
    week=None,
    inprogress=None,
    completed=None,
    pending=None,
    label=None,
    subtasks=False,
) -> list | None:
    """
    List all the tasks based on the filters.
    """
    order_by = "completed ASC, status ASC, priority DESC"
    where_clause = []
    if not subtasks:
        where_clause.append("parent_id ISNULL")
    if week:
        where_clause.append(
            "(completed > date('now', 'weekday 0', '-7 days') AND completed < date('now', 'weekday 1'))",
        )
    elif today:
        where_clause.append("(completed = date('now'))")
    if inprogress or completed or pending:
        clause = []
        if inprogress:
            clause.append("'In Progress'")
        if completed:
            clause.append("'Completed'")
        if pending:
            clause.append("'Pending'")
        where_clause.append("status in (" + ",".join(clause) + ")")
    else:
        clause = ["'In Progress'", "'Pending'"]
        where_clause.append("status in (" + ",".join(clause) + ")")

    if priority:
        where_clause.append(f"priority = {priority}")

    if label:
        where_clause.append(f"label = '{label}'")
    where_clause = "WHERE " + " AND ".join(where_clause)

    try:
        results = database.list_table(
            table=table,
            columns=[
                "id",
                "title",
                "parent_id",
                "status",
                "deadline",
                "priority",
                "label",
                "description",
                "subtasks",
            ],
            where_clause=where_clause,
            order_by=f"ORDER BY {order_by}",
        )
    except Exception as e:
        print(e)
        return None

    final_results = []
    for result in results:
        final_results.append(
            {
                "id": result[0],
                "title": result[1],
                "parent_id": result[2],
                "status": result[3],
                "deadline": (
                    result[4]
                    if str(result[4]) == "None"
                    else convert_to_console_date(result[4])
                ),
                "priority": result[5],
                "label": result[6] if result[6] else "None",
                "description": result[7],
                "subtasks": result[8],
            },
        )
    return final_results


def add_tasks(
    title: str,
    table="tasks",
    description=None,
    priority=None,
    today=False,
    week=False,
    deadline=None,
    inprogress=None,
    completed=None,
    pending=None,
    label=None,
    parent=None,
):
    """
    Add a task to the database.
    """
    columns = ["title"]
    values = [f"'{sanitize_text(title)}'"]
    if description:
        columns.append("description")
        values.append(f"'{sanitize_text(description)}'")
    if priority:
        columns.append("priority")
        values.append(str(priority))
    if today:
        columns.append("deadline")
        values.append("date('now')")
    elif week:
        columns.append("deadline")
        values.append("date('now', 'weekday 0')")
    elif deadline:
        columns.append("deadline")
        values.append(f"'{deadline}'")
    if inprogress:
        columns.append("status")
        values.append("'In Progress'")
    elif completed:
        columns.append("status")
        values.append("'Completed'")
    elif pending:
        columns.append("status")
        values.append("'Pending'")
    if label:
        columns.append("label")
        values.append(f"'{sanitize_text(label)}'")
    if parent:
        columns.append("parent_id")
        values.append(str(parent["id"]))
    try:
        database.insert_into_table(table, columns=columns, values=values)
    except Exception as e:
        print(e)
        return
    # Insert the record then increment the count of the parent task.
    if parent:
        database.update_table(
            table,
            {"subtasks": "subtasks + 1", "id": f"{parent['id']}"},
        )


def search_task(task_id, table: str) -> dict | None:
    """
    Search a task by its id.
    :param task_id:
    :return: task_details
    """
    try:
        task = database.list_table(
            table=table,
            columns=[
                "id",
                "title",
                "description",
                "status",
                "deadline",
                "priority",
                "label",
                "completed",
                "parent_id",
                "subtasks",
            ],
            where_clause=f"WHERE id = {task_id}",
        )
    except:
        generate_migration_error()
        return None

    task_details = {}
    if task:
        task = task[0]
        task_details = {
            "id": task[0],
            "title": task[1],
            "description": task[2],
            "status": task[3],
            "deadline": task[4],
            "priority": task[5],
            "label": task[6] if task[6] else "None",
            "completed": (task[7]),
            "parent_id": task[8],
            "subtasks": task[9],
        }
    return task_details


def get_subtasks(task_id: int, table: str):
    try:
        results = database.list_table(
            table=table,
            columns=[
                "id",
                "title",
                "status",
                "deadline",
                "priority",
                "label",
                "description",
                "subtasks",
                "parent_id",
            ],
            where_clause=f"WHERE parent_id = {task_id}",
            order_by="ORDER BY completed ASC, status ASC, priority DESC",
        )
    except:
        generate_migration_error()
        return None
    final_results = []
    for result in results:
        final_results.append(
            {
                "id": result[0],
                "title": result[1],
                "status": result[2],
                "deadline": (
                    result[3]
                    if str(result[3]) == "None"
                    else convert_to_console_date(result[3])
                ),
                "priority": result[4],
                "label": result[5] if result[5] else "None",
                "description": result[6],
                "subtasks": result[7],
                "parent_id": result[8],
            },
        )
    return final_results


def get_subtasks_recursive(task: dict, table: str):
    if task["subtasks"] == 0:
        return []

    final_results = [task]
    for child in get_subtasks(task["id"], table):
        final_results.append(child)
        final_results.extend(get_subtasks_recursive(child), table)
    return final_results


def update_task(updated_data: dict, table: str):
    """If marked as completed then set datetime as now else retain prev value"""

    if updated_data["deadline"] == "week":
        today = datetime.date.today()
        weekend = today + datetime.timedelta(days=6 - today.weekday())
        updated_data["deadline"] = str(weekend)
    elif updated_data["deadline"] == "today":
        updated_data["deadline"] = str(datetime.datetime.now().strftime("%Y-%m-%d"))
    elif updated_data["deadline"] not in [None, "None"]:
        updated_data["deadline"] = str(convert_to_db_date(updated_data["deadline"]))

    if updated_data["status"] == "Completed":
        updated_data["completed"] = str(datetime.datetime.now().strftime("%Y-%m-%d"))
    else:
        updated_data["completed"] = updated_data["deadline"]

    final_data = {}

    for key, value in updated_data.items():
        if value is None:
            continue

        if type(value) is str:
            updated_data[key] = f"'{sanitize_text(value)}'"

        final_data[key] = updated_data[key]

    try:
        database.update_table(table, final_data)
    except Exception as e:
        print(e)
        generate_migration_error()
        return
    if updated_data["subtasks"] != 0 and updated_data["status"] == "'Completed'":
        children = get_subtasks(updated_data["id"], table)
        for child in children:
            child["status"] = "Completed"
            try:
                update_task(child, table)
            except Exception:
                generate_migration_error()


def handle_delete(current_task: dict, table: str):
    """
    Delete a task from the database
    """
    database.delete_task(table, current_task["id"])
    children = database.list_table(
        table=table,
        columns=["id", "parent_id"],
        where_clause=f"WHERE parent_id = {current_task['id']}",
    )
    for child in children:
        handle_delete({"id": child[0], "parent_id": child[1]}, table)
    if current_task["parent_id"]:
        parent = search_task(current_task["parent_id"], table)
        if parent and parent["subtasks"] > 0:
            try:
                database.update_table(
                    "tasks",
                    {"subtasks": "subtasks - 1", "id": f"{current_task['parent_id']}"},
                )
            except:
                generate_migration_error()


def list_tables() -> list:
    """
    List all the tables in the database.
    """
    try:
        res = database.list_tables()
    except:
        generate_migration_error()
        return []

    result = []
    for table in res:
        if table[0] != "sqlite_sequence":
            result.append(table[0])
    return result


def add_table(table_name: str) -> bool:
    """
    Add a table to the database.
    """
    try:
        database.initialize(table_name)
        return True
    except Exception as e:
        print(e)
        return False


def delete_table(table_name: str) -> bool:
    """
    Delete a table from the database.
    """
    try:
        database.delete_table(table_name)
        return True

    except Exception as e:
        print(e)
        return False


def rename_table(old_name: str, new_name: str) -> bool:
    """
    Rename a table in the database.
    """
    try:
        database.rename_table(old_name, new_name)
        return True
    except Exception as e:
        print(e)
        return False



import json
import sys

from .__version__ import VERSION


def check_unicode_support():
    return sys.stdout.encoding.lower() in ("utf-8", "utf-16", "utf-32")


def initialize_config(
    path,
    theme="light",
    version=VERSION,
    pretty_tree=True,
    current_table="tasks",
    jira={},
    **kwargs,
):
    with open(path, "w+") as file:
        config = {
            "theme": theme,
            "version": version,
            "pretty_tree": pretty_tree,
            "current_table": current_table,
            "jira": jira,
        }
        json.dump(config, file, indent=4)
        return config


def update_config(path, config):
    initialize_config(path, **config)


def get_config(path):
    with open(path, "r") as file:
        return json.load(file)



import datetime

from click import echo
from click import style
from prompt_toolkit import prompt
from prompt_toolkit.completion import FuzzyWordCompleter
from prompt_toolkit.completion import ThreadedCompleter

from app import application


def convert_to_console_date(date_str, title=None):
    """
    Convert date from "YYYY-MM-DD" to "dd/mm/yyyy"
    """
    date_obj = datetime.datetime.strptime(date_str, "%Y-%m-%d")
    return date_obj.strftime("%d/%m/%Y")


def convert_to_db_date(date_str):
    # Convert date from "dd/mm/yyyy" to "YYYY-MM-DD"
    date_obj = datetime.datetime.strptime(date_str, "%d/%m/%Y")
    return date_obj.strftime("%Y-%m-%d")


def sanitize_text(text):
    return text.strip().replace("'", '"')


def fuzzy_search_task(table, completed=False) -> dict:
    all_tasks = application.list_tasks(table, subtasks=True, completed=completed)
    task_titles = [each_task["title"] for each_task in all_tasks]

    task_completer = ThreadedCompleter(FuzzyWordCompleter(task_titles))
    select_task_title = prompt(
        "Enter any part from title of the task: \n",
        completer=task_completer,
    )

    current_task = next(
        (
            each_task
            for each_task in all_tasks
            if each_task["title"] == select_task_title
        ),
        None,
    )
    return current_task


def generate_migration_error():
    echo(
        style(
            "Have You Run Migrations? Run 'devcord init --migrate' to run migrations",
            fg="red",
        ),
    )


def sanitize_table_name(table_name: str) -> (str, bool):
    for ch in table_name:
        if ch.isalnum() or ch == " " or ch == "_":
            continue
        else:
            return "", False

    return table_name.replace(" ", "_"), True


def check_table_exists(table_name: str) -> bool:
    table_name, ok = sanitize_table_name(table_name)
    if not ok:
        echo(
            style(
                "Error: Table name is not valid, please use only alphanumeric characters or underscores."
                + "Maybe you are not a developer?",
                fg="red",
            ),
        )
    return table_name in application.list_tables(), table_name



import app.database as db


def search_task(key):
    """
    Search a task by it's title.
    """
    try:
        task = db.list_table(
            table="jira",
            columns=[
                "id",
                "title",
                "description",
                "status",
                "deadline",
                "priority",
                "label",
                "completed",
                "parent_id",
                "subtasks",
            ],
            where_clause=f"WHERE title LIKE '{key}%'",
        )
    except Exception as e:
        print(e)
        return {}
    task_details = {}
    if task:
        task = task[0]
        task_details = {
            "id": task[0],
            "title": task[1],
            "description": task[2],
            "status": task[3],
            "deadline": task[4],
            "priority": task[5],
            "label": task[6] if task[6] else "None",
            "completed": (task[7]),
            "parent_id": task[8],
            "subtasks": task[9],
        }
    return task_details



import os
import unittest
from app.constants import path, db_path
from app.database import initialize
from app.application import list_tasks, add_tasks, search_task, get_subtasks, handle_delete, update_task
from datetime import datetime, timedelta

def create_db():
    os.makedirs(path, exist_ok=True)
    if os.path.exists(db_path):
        os.remove(db_path)
    initialize("tasks")

def fill_db():
    add_tasks(title = "Task 1", description="Description 1", priority=1)
    add_tasks(title = "Task 2", description="Description 2", priority=4, today=True)
    add_tasks(title = "Task 3", description="Description 3", priority=2, week=True)
    future_date = datetime.now() + timedelta(days=10)
    add_tasks(title = "Task 4", description="Description 4", priority=3, deadline=future_date.strftime("%Y-%m-%d"))
    add_tasks(title = "Task 5", description="Description 5", priority=5, completed=True)
    add_tasks(title = "Task 6", description="Description 6", inprogress=True)
    add_tasks(title = "Task 7", description="Description 7", pending=True)
    add_tasks(title = "Task 8", priority=3, deadline='2000-09-11')
    add_tasks(title='Child of task 1', parent={"id":1}, label='Label1')
    add_tasks(title='Child of child task 1', parent={"id": 9}, week=True)


class ListTasks(unittest.TestCase):
    def test_list_task_with_empty_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        # Test with no entries in table
        self.assertEqual(list_tasks(), [])

    def test_list_task_with_filled_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(), [
                {
                    "id": 8,
                    "title": "Task 8",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "11/09/2000",
                    "priority": 3,
                    "label": "None",
                    "description": "None",
                    "subtasks": 0,
                },
                {
                    "id": 2,
                    "title": "Task 2",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "31/08/2024",
                    "priority": 4,
                    "label": "None",
                    "description": "Description 2",
                    "subtasks": 0,
                },
                {
                    "id": 3,
                    "title": "Task 3",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "01/09/2024",
                    "priority": 2,
                    "label": "None",
                    "description": "Description 3",
                    "subtasks": 0,
                },
                {
                    "id": 4,
                    "title": "Task 4",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "10/09/2024",
                    "priority": 3,
                    "label": "None",
                    "description": "Description 4",
                    "subtasks": 0,
                },
                {
                    "id": 6,
                    "title": "Task 6",
                    "parent_id": None,
                    "status": "In Progress",
                    "deadline": "None",
                    "priority": 0,
                    "label": "None",
                    "description": "Description 6",
                    "subtasks": 0,
                },
                {
                    "id": 1,
                    "title": "Task 1",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "None",
                    "priority": 1,
                    "label": "None",
                    "description": "Description 1",
                    "subtasks": 1,
                },
                {
                    "id": 7,
                    "title": "Task 7",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "None",
                    "priority": 0,
                    "label": "None",
                    "description": "Description 7",
                    "subtasks": 0,
                },
            ]
        )

    def test_list_task_with_priority_and_today(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(priority=1, today=True), [])

    def test_list_task_with_priority(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(priority=1),[
                {
                    "id": 1,
                    "title": "Task 1",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "None",
                    "priority": 1,
                    "label": "None",
                    "description": "Description 1",
                    "subtasks": 1,
                }
            ])

    def test_list_task_with_today(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(today=True), [
                {
                    "id": 2,
                    "title": "Task 2",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "31/08/2024",
                    "priority": 4,
                    "label": "None",
                    "description": "Description 2",
                    "subtasks": 0,
                }
            ]
        )

    def test_list_task_with_week(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(week=True, subtasks=True), [
                {
                    "id": 2,
                    "title": "Task 2",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "31/08/2024",
                    "priority": 4,
                    "label": "None",
                    "description": "Description 2",
                    "subtasks": 0,
                },
                {
                    "id": 3,
                    "title": "Task 3",
                    "parent_id": None,
                    "status": "Pending",
                    "deadline": "01/09/2024",
                    "priority": 2,
                    "label": "None",
                    "description": "Description 3",
                    "subtasks": 0,
                },
                {
                    "id": 10,
                    "title": "Child of child task 1",
                    "parent_id": 9,
                    "status": "Pending",
                    "deadline": "01/09/2024",
                    "priority": 0,
                    "label": "None",
                    "description": "None",
                    "subtasks": 0,
                },
            ]
                         )

    def test_list_task_with_status(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(inprogress=True, completed=True), [
                {
                    "id": 5,
                    "title": "Task 5",
                    "parent_id": None,
                    "status": "Completed",
                    "deadline": "None",
                    "priority": 5,
                    "label": "None",
                    "description": "Description 5",
                    "subtasks": 0,
                },
                {
                    "id": 6,
                    "title": "Task 6",
                    "parent_id": None,
                    "status": "In Progress",
                    "deadline": "None",
                    "priority": 0,
                    "label": "None",
                    "description": "Description 6",
                    "subtasks": 0,
                },
            ]
        )

    def test_list_task_with_label(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(list_tasks(label='Label1', subtasks=True), [
                {
                    "id": 9,
                    "title": "Child of task 1",
                    "parent_id": 1,
                    "status": "Pending",
                    "deadline": "None",
                    "priority": 0,
                    "label": "Label1",
                    "description": "None",
                    "subtasks": 1,
                }
            ]
        )

class AddTasks(unittest.TestCase):
    def test_adding_task_with_single_quotes_in_description(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()

        self.assertEqual(add_tasks(title='Dummy Task', description="'Description'"), None)

    def test_adding_task_with_special_characters(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()

        self.assertEqual(add_tasks(title='Dummy Task', description="$ / \ "), None)

class SearchTask(unittest.TestCase):
    def test_search_task_with_empty_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()

        self.assertEqual(search_task(1, "tasks"), {})

    def test_search_task_with_filled_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(search_task(1, "tasks"), {
                "id": 1,
                "title": "Task 1",
                "description": "Description 1",
                "status": "Pending",
                "deadline": "None",
                "priority": 1,
                "label": "None",
                "completed": "None",
                "parent_id": None,
                "subtasks": 1,
            }
        )

    def test_search_task_with_invalid_id(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(search_task(100, "tasks"), {})

class GetSubtasks(unittest.TestCase):
    def test_get_subtasks_with_empty_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()

        self.assertEqual(get_subtasks(1, "tasks"), [])

    def test_get_subtasks_with_filled_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(
            get_subtasks(1, "tasks"),
            [
                {
                    'id': 9,
                    'title': 'Child of task 1',
                    'status': 'Pending',
                    'deadline': 'None',
                    'priority': 0,
                    'label': 'Label1',
                    'description': 'None',
                    'subtasks': 1,
                    'parent_id': 1
                }
            ]
        )

    def test_get_subtasks_with_invalid_id(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()

        self.assertEqual(get_subtasks(100, "tasks"), [])

class HandleDelete(unittest.TestCase):
    def test_with_filled_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()
        # test with valid id
        handle_delete({"id": 1, "parent_id": None}, "tasks")
        self.assertEqual(list_tasks(), [
            {
                "id": 8,
                "title": "Task 8",
                "parent_id": None,
                "status": "Pending",
                "deadline": "11/09/2000",
                "priority": 3,
                "label": "None",
                "description": "None",
                "subtasks": 0,
            },
            {
                "id": 2,
                "title": "Task 2",
                "parent_id": None,
                "status": "Pending",
                "deadline": "31/08/2024",
                "priority": 4,
                "label": "None",
                "description": "Description 2",
                "subtasks": 0,
            },
            {
                "id": 3,
                "title": "Task 3",
                "parent_id": None,
                "status": "Pending",
                "deadline": "01/09/2024",
                "priority": 2,
                "label": "None",
                "description": "Description 3",
                "subtasks": 0,
            },
            {
                "id": 4,
                "title": "Task 4",
                "parent_id": None,
                "status": "Pending",
                "deadline": "10/09/2024",
                "priority": 3,
                "label": "None",
                "description": "Description 4",
                "subtasks": 0,
            },
            {
                "id": 6,
                "title": "Task 6",
                "parent_id": None,
                "status": "In Progress",
                "deadline": "None",
                "priority": 0,
                "label": "None",
                "description": "Description 6",
                "subtasks": 0,
            },
            {
                "id": 7,
                "title": "Task 7",
                "parent_id": None,
                "status": "Pending",
                "deadline": "None",
                "priority": 0,
                "label": "None",
                "description": "Description 7",
                "subtasks": 0,
            },
        ])
        self.assertEqual(list_tasks(week=True, subtasks=True), [
            {
                "id": 2,
                "title": "Task 2",
                "parent_id": None,
                "status": "Pending",
                "deadline": "31/08/2024",
                "priority": 4,
                "label": "None",
                "description": "Description 2",
                "subtasks": 0,
            },
            {
                "id": 3,
                "title": "Task 3",
                "parent_id": None,
                "status": "Pending",
                "deadline": "01/09/2024",
                "priority": 2,
                "label": "None",
                "description": "Description 3",
                "subtasks": 0,
            },
        ])

    def test_subtask_deletion(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()
        # test with subtask id
        handle_delete({"id": 9, "parent_id": 1}, "tasks")
        self.assertEqual(search_task(1, "tasks"), {
                "id": 1,
                "title": "Task 1",
                "description": "Description 1",
                "status": "Pending",
                "deadline": "None",
                "priority": 1,
                "label": "None",
                "completed": "None",
                "parent_id": None,
                "subtasks": 0,
            })


class HandleModify(unittest.TestCase):
    def test_cascading_effect_with_filled_db(self):
        # check if DEBUG is set to True
        self.assertTrue(os.environ.get("DEBUG", "") == "True")
        # set test environment
        create_db()
        fill_db()
        task = search_task(1, "tasks")
        task['status'] = 'Completed'
        update_task(task, "tasks")   # cascading update
        self.assertEqual(search_task(1, "tasks"), {'id': 1, 'title': 'Task 1', 'description': 'Description 1', 'status': 'Completed', 'deadline': 'None', 'priority': 1, 'label': 'None', 'completed': '2024-08-31', 'parent_id': None, 'subtasks': 1})
        self.assertEqual(get_subtasks(1, "tasks"), [{'id': 9, 'title': 'Child of task 1', 'status': 'Completed', 'deadline': 'None', 'priority': 0, 'label': 'Label1', 'description': 'None', 'subtasks': 1, 'parent_id': 1}])
        self.assertEqual(get_subtasks(9, "tasks"), [{'id': 10, 'title': 'Child of child task 1', 'status': 'Completed', 'deadline': '01/09/2024', 'priority': 0, 'label': 'None', 'description': 'None', 'subtasks': 0, 'parent_id': 9}])


if __name__ == '__main__':
    unittest.main()



import click
import keyring

from app.config import update_config
from app.constants import config_path


def set_jira_config(config):
    if config.get("jira", None) is None:
        config["jira"] = {}
        config = set_organization_url(config=config)
        config = set_organization_email(config=config)
        set_token()

    return config


def set_organization_url(config):
    config["jira"]["url"] = click.prompt(
        "Enter Jira URL for API (Ex: 'https://<enterprise>jira.atlassian.net/')",
        type=str,
    )
    update_config(config_path, config)
    return config


def set_organization_email(config):
    config["jira"]["email"] = click.prompt("Enter oragnizational Email", type=str)

    update_config(config_path, config)
    return config


def set_token():
    token = click.prompt("Enter private API token", type=str)
    keyring.set_password("devcord", "token", token)



import os

import click

from app import application
from app.config import check_unicode_support
from app.config import get_config
from app.config import initialize_config
from app.config import update_config
from app.console import print_legend
from app.console import print_tables
from app.console import print_tasks
from app.constants import config_path
from app.constants import db_path
from app.constants import path
from app.database import initialize
from app.migrations import update_version
from app.utility import check_table_exists
from app.utility import convert_to_db_date
from app.utility import fuzzy_search_task
from jira.application import update_issues
from jira.console import set_jira_config
from jira.console import set_organization_email
from jira.console import set_organization_url
from jira.console import set_token

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.pass_context
def cli(ctx):
    """
    Devcord is a CLI tool for developers to help them with their daily tasks.
    """
    ctx.ensure_object(dict)
    if not path:
        click.echo(
            click.style(
                "Error: Could not find the path to the database, raise an issue with the developers.",
                fg="red",
            ),
        )
        ctx.abort()
        return

    # db_path as well as config_path can only be set if path is not None
    if not os.path.exists(db_path):
        os.makedirs(path, exist_ok=True)
        initialize("tasks")
    if not os.path.exists(config_path):
        ctx.obj["config"] = initialize_config(config_path)
    else:
        ctx.obj["config"] = get_config(config_path)

    ctx.obj["config"]["unicode"] = check_unicode_support()
    update_version(ctx.obj["config"])


@cli.command()
@click.pass_context
@click.option("-l", "--list", is_flag=True, help="List all the tasks")
@click.option("-a", "--add", help="Add a new task", type=str)
@click.option("-d", "--desc", is_flag=True, help="Add a description to a task")
@click.option("-p", "--priority", help="Set the priority of a task", type=int)
@click.option("-t", "--today", is_flag=True, help="Perform for all the tasks for today")
@click.option(
    "-w",
    "--week",
    is_flag=True,
    help="Perform for all the tasks for this week",
)
@click.option(
    "-dd",
    "--deadline",
    help='Set the deadline of a task, date format: "dd/mm/yyyy/"',
    type=str,
)
@click.option(
    "-i",
    "--inprogress",
    is_flag=True,
    help="Perform for all the tasks that are in progress",
)
@click.option(
    "-c",
    "--completed",
    is_flag=True,
    help="Perform for all the tasks that are completed",
)
@click.option(
    "-pd",
    "--pending",
    is_flag=True,
    help="Perform for all the tasks that are pending",
)
@click.option(
    "-lb",
    "--label",
    help="Perform for all the tasks with a specific label",
    type=str,
)
@click.option("-o", "--output", help="Specify Output Format", type=str)
@click.option("--path", help="Specify Output File", type=str)
@click.option("-st", "--subtask", is_flag=True, help="List or add subtasks")
@click.option("-tb", "--table", help="Specify Table", type=str)
def tasks(
    ctx,
    list=None,
    add=None,
    desc=None,
    priority=None,
    today=False,
    week=False,
    deadline=None,
    inprogress=None,
    completed=None,
    pending=None,
    label=None,
    output=None,
    path=None,
    subtask=False,
    table=None,
):
    """
    Create and List tasks.
    """

    if not table:
        table = ctx.obj["config"].get("current_table", "tasks")

    if not add and not list:
        click.echo(
            click.style(
                "Error: Please specify an action.",
                fg="red",
            ),
        )
        return

    if deadline:
        try:
            deadline = convert_to_db_date(deadline)
        except ValueError:
            click.echo(
                click.style(
                    'Error: Invalid date format, please use "dd/mm/yyyy".',
                    fg="red",
                ),
            )
            click.echo('Example: "01/01/2020"')
            return

    if list:

        task_list = application.list_tasks(
            table=table,
            priority=priority,
            today=today,
            week=week,
            inprogress=inprogress,
            completed=completed,
            pending=pending,
            label=label,
            subtasks=subtask,
        )

        if (
            priority is not None
            or today
            or week
            or inprogress
            or completed
            or pending
            or label
        ):
            subtask = False

        if task_list:
            print_tasks(
                task_list,
                output,
                path,
                ctx.obj["config"]["unicode"] is False,
                subtask,
                table,
                pretty_tree=ctx.obj["config"].get("pretty_tree", True),
            )
        else:
            click.echo(
                click.style(
                    "Info: No tasks found.",
                    fg="yellow",
                ),
            )

    elif add:
        parent = None
        description = "No given description"
        if desc:
            description = click.edit()
        if subtask:
            parent = fuzzy_search_task(table)
            if parent is None:
                click.echo(
                    click.style(
                        "Error: Parent task does not exist.",
                        fg="red",
                    ),
                )
                return

        application.add_tasks(
            add,
            table,
            description,
            priority,
            today,
            week,
            deadline,
            inprogress,
            completed,
            pending,
            label,
            parent,
        )


@cli.command()
@click.pass_context
@click.option(
    "-d",
    "--desc",
    help="View and edit description of the task",
    is_flag=True,
)
@click.option(
    "-i",
    "--inprogress",
    is_flag=True,
    help="Mark Task As In Progress",
)
@click.option(
    "-c",
    "--completed",
    is_flag=True,
    help="Mark Task As Completed",
)
@click.option(
    "-pd",
    "--pending",
    is_flag=True,
    help="Mark Task As Pending",
)
@click.option(
    "-st",
    "--subtasks",
    is_flag=True,
    help="List All Subtask Of Task",
)
@click.option(
    "-w",
    "--week",
    is_flag=True,
    help="Change deadline to this week",
)
@click.option("-t", "--today", is_flag=True, help="Change deadline to today")
@click.option("-dl", "--delete", is_flag=True, help="Delete task")
@click.option("-n", "--name", help="Change the name of the task", type=str)
@click.option("-p", "--priority", help="Change the priority of the task", type=int)
@click.option("-dd", "--deadline", help="Change the deadline of the task", type=str)
@click.option("-lb", "--label", help="Change the label of the task", type=str)
@click.option("-ar", "--archive", is_flag=True, help="Edit Completed the task")
@click.option("-tb", "--table", help="Specify Table", type=str)
def task(
    ctx,
    desc=None,
    inprogress=None,
    completed=None,
    pending=None,
    subtasks=None,
    week=False,
    today=False,
    delete=None,
    name=None,
    priority=None,
    deadline=None,
    label=None,
    archive=False,
    table=None,
):
    """
    Modify a specific task.
    """

    if table is None:
        table = ctx.obj["config"].get("current_table", "tasks")

    current_task = fuzzy_search_task(table, archive)
    if current_task is None:
        click.echo(
            click.style(
                "Error: Task does not exist.",
                fg="red",
            ),
        )
        return

    if inprogress:
        current_task["status"] = "In Progress"
    elif pending:
        current_task["status"] = "Pending"
    elif completed:
        current_task["status"] = "Completed"

    if name:
        current_task["title"] = name
    if priority:
        current_task["priority"] = priority
    if label:
        current_task["label"] = label
    if week:
        current_task["deadline"] = "week"
    elif today:
        current_task["deadline"] = "today"
    elif deadline:
        try:
            convert_to_db_date(deadline)
            current_task["deadline"] = deadline
        except ValueError:
            click.echo(
                click.style(
                    'Error: Invalid date format, please use "dd/mm/yyyy".',
                    fg="red",
                ),
            )
            return

    if subtasks:
        task_list = application.get_subtasks_recursive(current_task, table)
        if task_list:
            temp = current_task.copy()
            temp["parent_id"] = (
                -1
            )  # For subtasks, the parent of the parent node is irrelevant
            # Parent_id cannot be -1, therefore functions ahead can recognize this node as root.
            task_list.append(temp)
            print_tasks(
                tasks=task_list,
                plain=ctx.obj["config"]["unicode"] is False,
                subtasks=subtasks,
                table_name=table,
                pretty_tree=ctx.obj["config"].get("pretty_tree", True),
            )
            return

    elif desc:
        description = "No given description"
        if current_task["description"]:
            description = current_task["description"]
        current_task["description"] = click.edit(description)

    application.update_task(current_task, table)

    if delete:
        application.handle_delete(current_task, table)
        return


@cli.command()
@click.pass_context
@click.option("-l", "--list", is_flag=True, help="List all the tables")
@click.option("-a", "--add", help="Add a new table", type=str)
@click.option("-sl", "--select", help="Select a table", type=str)
@click.option("-dl", "--delete", help="Delete a table", type=str)
@click.option("-n", "--name", help="Change the name of the table", type=str)
def tables(ctx, list=None, add=None, select=None, delete=None, name=None):
    """
    Use multiple tables for segregating your tasks.
    """
    if list:
        table_list = application.list_tables()
        print_tables(table_list, ctx.obj["config"].get("current_table", "tasks"))

    elif add:
        exists, add = check_table_exists(add)
        if exists:
            click.echo(
                click.style(
                    "Error: Table already exists.",
                    fg="red",
                ),
            )
            return

        ok = application.add_table(add)

        if ok:
            click.echo(
                click.style(
                    "Success: ",
                    fg="green",
                )
                + "Table stored as "
                + add,
            )

    elif select:
        exists, select = check_table_exists(select)
        if not exists:
            click.echo(
                click.style(
                    "Error: Table does not exist.",
                    fg="red",
                ),
            )
            return

        ctx.obj["config"]["current_table"] = select
        update_config(config_path, ctx.obj["config"])

        click.echo(
            click.style(
                "Success: ",
                fg="green",
            )
            + "Table selected "
            + select,
        )

    elif delete:
        exists, delete = check_table_exists(delete)
        if not exists:
            click.echo(
                click.style(
                    "Error: Table does not exist.",
                    fg="red",
                ),
            )
            return
        if application.list_tables() == 1:
            click.echo(
                click.style(
                    "Error: Cannot delete the only table.",
                    fg="red",
                ),
            )
            return

        ok = application.delete_table(delete)

        if ok:
            click.echo(
                click.style(
                    "Success: ",
                    fg="green",
                )
                + "Table deleted: "
                + delete,
            )

    elif name:
        exists, name = check_table_exists(name)
        if not exists:
            click.echo(
                click.style(
                    "Error: Table does not exist.",
                    fg="red",
                ),
            )
            return

        new_name = click.prompt(
            "Enter new name for the table",
            default=name,
            show_default=False,
        )

        ok = application.rename_table(name, new_name)

        if ok:
            click.echo(
                click.style(
                    "Success: ",
                    fg="green",
                )
                + "Table renamed from "
                + click.style(name, fg="yellow")
                + " to: "
                + click.style(new_name, fg="yellow"),
            )


@cli.command()
@click.pass_context
@click.option(
    "--sync",
    is_flag=True,
    help="Sync with your tasks on Jira",
)
@click.option(
    "--token",
    is_flag=True,
    help="Set private API token",
)
@click.option(
    "--url",
    is_flag=True,
    help="Set organization Jira url",
)
@click.option(
    "--email",
    is_flag=True,
    help="Set oraganization email",
)
def jira(ctx, sync=False, token=False, url=False, email=False):
    """
    Import your work from Jira
    """
    set_jira_config(ctx.obj["config"])

    if token:
        set_token()

    if url:
        set_organization_url(ctx.obj["config"])

    if email:
        set_organization_email(ctx.obj["config"])

    if sync:
        exists, _ = check_table_exists("jira")
        if not exists:
            application.add_table("jira")

        update_issues(
            ctx.obj["config"]["jira"]["url"],
            ctx.obj["config"]["jira"]["email"],
            "jira",
        )
        click.echo("Issues synced")


@cli.command()
@click.pass_context
def legend(ctx):
    """
    Show the legend for all special characters
    """
    if ctx.obj["config"]["unicode"]:
        print_legend()
    else:
        click.echo(
            click.style(
                "Info: Unicode is disabled, legend not required",
                fg="yellow",
            ),
        )


@cli.command()
@click.pass_context
@click.option("--migrate", is_flag=True, help="Migrate database")
@click.option("--pretty_tree", help="Change the name of the task", type=bool)
def init(ctx, migrate=False, pretty_tree=None):
    """
    Run after every install
    """
    if migrate:
        update_version(ctx.obj["config"])
    if pretty_tree is not None:
        ctx.obj["config"]["pretty_tree"] = pretty_tree
        update_config(config_path, ctx.obj["config"])
        click.echo(
            click.style(
                "Info: Pretty Tree setting updated",
                fg="yellow",
            ),
        )



import json
import os

import PrettyPrint
from click import echo
from click import style
from PrettyPrint import PrettyPrintTree
from rich.console import Console
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from . import console_helper


def get_table(tasks, plain=False):
    table = Table(title="Tasks", highlight=True, leading=True)
    table.add_column("Priority", justify="center", style="white")
    table.add_column("Task", justify="left", style="white")
    table.add_column("Status", justify="center", style="white")
    table.add_column("Deadline", justify="center", style="white")
    table.add_column("Label", justify="center", style="white")
    table.add_column("Properties", justify="center", style="white")
    text_style = Style(color="#FFFFFF")
    bold_text_style = Style(color="#FFFFFF", bold=True)
    none_style = Style(color="magenta")
    for task in tasks:
        properties = []
        if task["description"] and task["description"] not in [
            "None",
            "No given description",
        ]:
            if plain:
                properties.append("Description")
            else:
                properties.append("►")
        if task["subtasks"] > 0:
            if plain:
                properties.append("Subtasks")
            else:
                properties.append("|☰")

        table.add_row(
            (
                f"[{console_helper.get_priority_color(task['priority'])}]{task['priority']}"
                if not plain
                else f"[{text_style}]{task['priority']}"
            ),
            f'[{text_style}]{task["title"]}',
            f'[{console_helper.get_status_color(task["status"])}][italic]{task["status"]}',
            task["deadline"],
            f'[{bold_text_style if task["label"] != "None" else none_style}]{task["label"]}',
            f"[{text_style}]{','.join(properties)}",
        )
    return table


def print_tree(tasks, table_name="Tasks", pretty_tree=True):

    # TODO use the moved out the logic of forming the tree and have to only print it here
    def parse_subtask_string(task):
        if pretty_tree == False:
            text = Text()
            text.append(f" {task['title']}\n", style="italic bold")
            text.append("Priority: ", style="dim")
            text.append(
                f"{task['priority']} ",
                style=f"{console_helper.get_priority_color(task['priority'])}",
            )
            text.append(f"| {task['deadline']} | ", style="dim")
            text.append(
                f"{task['status']} ",
                style=console_helper.get_status_color(task["status"]) + " dim",
            )
            text.append(f"| {task['label']}\n", style="dim")
            # return f" [italic]{task['title']}\n[bold]Priority: [white]{task['priority']} | {task['deadline']}| {task['status']} | {task['label']}\n"
            return text
        else:
            return f"{task['title']}\nPriority: | {task['deadline']}| {task['status']} | {task['label']}"

    root = Tree("\n" + table_name + "\n", style="bold")
    root.guide_style = "bold blue"
    existing_node = {}

    for task in tasks:

        if task["parent_id"] in (None, -1):
            if existing_node.get(task["id"], False):
                existing_node[task["id"]].label = parse_subtask_string(task)
            else:
                existing_node[task["id"]] = Tree(parse_subtask_string(task))
            if task["parent_id"] == -1:
                root = existing_node[task["id"]]
            else:
                existing_node[task["id"]].guide_style = "yellow"
                root.children.append(existing_node[task["id"]])
        else:

            if task["id"] in existing_node:
                existing_node[task["id"]].label = parse_subtask_string(task)
            else:
                existing_node[task["id"]] = Tree(parse_subtask_string(task))

            if task["parent_id"] not in existing_node:
                parent_task = Tree("temp")  # temparary node for adding child
                existing_node[task["parent_id"]] = parent_task

            existing_node[task["id"]].guide_style = "yellow"
            existing_node[task["parent_id"]].children.append(existing_node[task["id"]])

    if pretty_tree:
        pt = PrettyPrintTree(
            lambda x: x.children,
            lambda x: x.label,
            border=False,
            orientation=PrettyPrintTree.Horizontal,
        )
        pt(root)
    else:
        return root


def print_tasks(
    tasks,
    output=None,
    path=None,
    plain=False,
    subtasks=False,
    table_name="Tasks",
    pretty_tree=True,
):
    if subtasks:
        if output == "json":
            result = console_helper.treeify(tasks)
        elif output or path or pretty_tree == False:
            # passing result ahead as rich can print its own object
            result = print_tree(tasks, table_name, False)
        else:
            print_tree(tasks, table_name)
            return

    # Setting where the output will be sent
    file = None
    if path:
        path = path.strip()
        if path[0] != "/":  # If path is not absolute
            path = os.path.join(os.getcwd(), path)

        if not console_helper.sanitize_path(path):
            return

        file = open(path, "w+")
        console = Console(file=file)
    else:
        console = Console()

    # Handling output format
    if path:
        plain = True
    if output == "json":
        if subtasks:
            result = json.dumps(tasks, indent=4)
        console.print_json(result)

    elif output == "text":
        if subtasks:
            console.print(result)
        else:
            console.print(f"[bold]{table_name}")
            console.rule()
            index = 1
            for task in tasks:
                console.print(
                    f"{index}) Title: {task['title']}\n- Description: {task['description']}\n- Deadline: {task['deadline']}\n",
                )
                index += 1

    else:
        if not subtasks:
            result = get_table(tasks, plain)
        console.print(result)

    if file:
        file.close()


def print_legend():
    text_style = Style(color="#FFFFFF")
    console = Console()
    table = Table(title="Priority Legend", highlight=True, leading=True)
    table.add_column("Unicode Character", justify="center", style="white")
    table.add_column("Value", justify="center", style="white")
    table.add_row(
        f"[{console_helper.get_priority_color(0)}]●",
        f"[{text_style}]No Priority Level",
    )
    for i in range(1, 6):
        table.add_row(
            f"[{console_helper.get_priority_color(i)}]●",
            f"[{text_style}]Priority level: {i}",
        )
    console.print(table)
    console.rule()
    table = Table(title="Task Properties Legend", highlight=True, leading=True)
    table.add_column("Unicode Character", justify="center", style="white")
    table.add_column("Value", justify="center", style="white")
    table.add_row("[#FFFFFF]►", f"[{text_style}]Has description")
    table.add_row("[#FFFFFF]|☰", f"[{text_style}]Has subtasks")
    console.print(table)


def print_tables(tables, current_table):
    console = Console()
    table = Table(title="Task Lists", highlight=True, leading=True)
    table.add_column("List Name", justify="center", style="white")
    for table_name in tables:
        if table_name == current_table:
            table.add_row(f"[#FFFFFF][italic]-> {table_name}")
        else:
            table.add_row(f"[#FFFFFF]{table_name}")
    console.print(table)



import os
import platform


if platform.system() == "Windows":
    home_word_list = ["USERPROFILE", "HOMEDRIVE", "HOMEPATH"]
    for home_word in home_word_list:
        if home_word in os.environ:
            path = os.path.join(os.environ[home_word], ".devcord")
            break
    else:
        path = None
else:
    # Path value can be None as there is a check in devcord.py
    path = os.path.join(os.getenv("HOME"), ".devcord")

if path:
    if os.environ.get("DEBUG", "") == "True":
        db_path = os.path.join(path, "test.db")
        config_path = os.path.join(path, "test_config.json")
    else:
        db_path = os.path.join(path, "data.db")
        config_path = os.path.join(path, "config.json")
else:
    db_path = None
    config_path = None
